# -*- coding: utf-8 -*-
# Module: lwMCMC
# Author: Daniel Ryan Furman <dryanfurman@gmail.com>
# License: MIT
# Release: lwMCMC 0.1.0
# Last modified : 25/10/2020
# Github: https://github.com/daniel-furman/lwMCMC

import numpy as np
import matplotlib.pyplot as plt

class lwMCMC(object):
    
    """ MCMC chain using a lightweight implementation of the Metropolis
    Hastings search algorithm, an object-oriented class with the 
    following inputs:
        
    * log_likelihood - function returning the log of the likelihood
       p(data|theta), pre-defined (see examples). The function should
       take two variables (data, theta) and return a single value
       log(p(data | theta)).

    * data is the fixed input data in the log_likelihood form

    * theta is a list or array with the starting parameter values for the
        Marcov chain.

    * step_size is a list or array with the step size in each dimension of
        theta."""
    
    def __init__(self, log_likelihood, data, theta, step_size, names=None,
                 seed=2145):
        self.log_likelihood = log_likelihood
        self.data = data
        self.theta = np.array(theta)
        self.nparams = len(theta)
        self.step_size = np.array(step_size)
        self.rng = np.random.RandomState(seed)
        self.naccept = 0
        self.current_loglike = log_likelihood(self.data, self.theta)
        self.samples = []
        if names is None:
            names = ["Paramter {:d}".format(k+1) for k in range(self.nparams)]
        self.names = names            

    def step(self, save=True):
        """Take a step in the MCMC chain (a single step)"""
        new_theta = self.theta + self.step_size * self.rng.normal(
            size=len(self.step_size))
        new_loglike = self.log_likelihood(self.data, new_theta)
        diff = new_loglike - self.current_loglike

        if diff >= 0:
            take_step = True
        else:
            take_step = self.rng.uniform() < np.exp(diff)

        if take_step:
            self.current_loglike = new_loglike
            self.theta = new_theta

        if save:
            self.samples.append(self.theta)
            if take_step: 
                self.naccept += 1

    def burn(self, nburn):
        """Number of burns (results not saved)"""
        for i in range(nburn):
            self.step(save=False)

    def run(self, nsteps):
        """Take nsteps steps (results saved)"""
        for i in range(nsteps):
            self.step()

    def accept_fraction(self):
        """Returns the fraction of candidate steps that were accpeted
        so far."""
        if len(self.samples) > 0:
            return float(self.naccept) / len(self.samples)
        else:
            return 0.
        
    def clear(self, step_size=None, theta=None):
        """Clear the list of stored samples from any runs so far.
        Optional change step_size or theta here."""
        if step_size is not None:
            assert len(step_size) == self.nparams
            self.step_size = np.array(step_size)
        if theta is not None:
            assert len(theta) == self.nparams
            self.theta = np.array(theta)
            self.current_loglike = self.log_likelihood(self.data, self.theta)
        self.samples = []
        self.naccept = 0
        
    def get_samples(self):
        """Return the sampled theta values at each step in the chain as a
        2d numpy array."""
        return np.array(self.samples)
        
    def plot_hist(self):
        """Plot a histogram of the sample values for each parameter in the
        theta vector."""
        all_samples = self.get_samples()
        for k in range(self.nparams):
            theta_k = all_samples[:,k]
            plt.hist(theta_k, bins=100)
            plt.xlabel(self.names[k])
            plt.ylabel("N Samples")
            plt.show()
        
    def plot_samples(self):
        """Plot the sample values over the course of the chain so far."""
        all_samples = self.get_samples()
        for k in range(self.nparams):
            theta_k = all_samples[:,k]
            plt.plot(range(len(theta_k)), theta_k)
            plt.xlabel("Step in chain")
            plt.ylabel(self.names[k])
            plt.show()

    def calculate_mean(self, weight=None):
        """Calculate the mean of each parameter according to the samples
        taken so far. Optionally, provide a weight array to weight
        the samples."""
        return np.average(self.get_samples(), axis=0, weights=weight)
    
    def calculate_cov(self, weight=None):
        """Calculate the covariance matrix of the parameters according to
        the samples taken so far. Optionally, provide a weight array to 
        weight the samples."""
        return np.cov(self.get_samples(), rowvar=False, aweights=weight)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Feb  3 23:17:41 2021

@author: danielfurman
"""

import os
import sys
from setuptools import setup, find_packages

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

with open('README.md') as f:
    readme = f.read()

setup(
    name="lwMCMC",
    version="0.1",
    author="Daniel Ryan Furman",
    author_email="dryanfurman@gmail.com",
    description=("Parameter space sampling with lightweight MCMC powered by NumPy and Metropolis Hastings"),
    long_description="See documentation at https://github.com/daniel-furman/lwMCMC",
    license="MIT",
    keywords="bayesian-inference machine-learning statistical-modeling",
    url="https://github.com/daniel-furman/lwMCMC",
    packages=find_packages(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Utilities",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering"
        ],

)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jan  7 10:36:32 2021

@author: danielfurman
"""

import numpy as np
from pytest import fixture

@fixture
def op():
    from lwMCMC._main import lwMCMC as MCMC
    return MCMC(gauss_log_likelihood, data, theta, step_size,
                names=('$x$','$y$','$z$'))


# Unit tests for general purpose utility, MCMC.
# Let's test this with a 3D Gaussian.

def gauss_log_likelihood(data, theta):
    """Comput a Gaussian log(likelihood) for a trial value theta = (x,y,z)

    Here, the "data" are just the fixed values (x0,y0,z0) where the Gaussian
    is centered.

    The theta parameters are the MCMC's trial of x,y,z.  When they match
    (x0,y0,z0), that will correspond to the maximum likelihood.  For other
    values of (x,y,z), the likelihood is smaller.
    """
    x, y, z = theta
    x0, y0, z0, sigma = data
    # likelihood = exp(-((x-x0)^2 + (y-y0)^2 + (z-z0)^2) / (2 sigma^2))
    # log(likelihood) = -((x-x0)^2 + (y-y0)^2 + (z-z0)^2) / (2 sigma^2))
    return -((x-x0)**2 + (y-y0)**2 + (z-z0)**2) / (2 * sigma**2)

x0, y0, z0 = 12, 21, 30    # The peak location is at (12, 21, 30)
sigma = 0.5                # Uncertainty on each is 0.5
data = (x0, y0, z0, sigma) # The "data" for this toy model

assert gauss_log_likelihood(data, (x0,y0,z0)) == 0.  # Peak has a value of 0
assert gauss_log_likelihood(data, (x0-0.01, y0+0.01, z0-0.01)) < 0.
  # Any other point, no matter how close, is < 0.

# Use an initial guess of (0,0,0)
theta = np.zeros(3)
#print('log likelihood at (0,0,0) = ',gauss_log_likelihood(data, theta))

# Note that the likelihood underflows to 0.
#print('likelihood at (0,0,0) = ',np.exp(gauss_log_likelihood(data, theta)))
# But the log likelihood is fine.

assert np.isfinite(gauss_log_likelihood(data, theta))

# Make the mcmc object with this likelihood function
step_size = (0.5,0.5,0.5)  # 1 sigma in each direction

def testers(op):

    # Run for 1000 steps
    op.run(1000)

    # Plot what we have from that.
    #print('First 1000 steps:')
    #op.plot_samples()

    # The initial samples are clearly biased.
    # But by 300 or so, it seems to have settled into a steady state behavior.
    # This means that around 500 steps is a safe value for the burn in.

    # Start over from our initial guess and burn 1000 before running 100000 steps
    op.clear(theta=theta)
    op.burn(1000)
    op.run(100000)

    # Now we shouldn't see any evidence of the initial ramp.
    #print('After burn of 1000, next 100,000 steps:')
    #op.plot_samples()

    # We'd like an acceptance rate near 50%.
    #print('Acceptance rate is ', op.accept_fraction())
    #print()
    assert 0.2 < op.accept_fraction() < 0.8

    # The mean should be pretty close to the actual maximum likelihood.
    mean = op.calculate_mean()
    #print('Mean values of (x,y,z) = ', mean)
    #print('loglike at mean = ', gauss_log_likelihood(data, mean))
    #print()

    assert np.allclose(mean, (x0,y0,z0), rtol=1.e-3)

    # The covariance matrix is less well estimated, but still well witin 10%
        # with only 10,000 steps
    cov = op.calculate_cov()
    #print('Uncertainties of (x,y,z) = ', np.sqrt(cov.diagonal()))
    #print('Full covariance matrix = \n', cov)
    #print()
    assert np.allclose(cov.diagonal(), sigma**2, rtol=0.1)


    # Show the one-d distributions for each of (x,y,z)
    # These should look fairly smooth.
    #op.plot_hist()

    # Finally, check the weighted average options
    # Let's say you learn that y - x > 10.
    samples = op.get_samples()
    x_samples, y_samples, z_samples = samples.T  # extract by columns.
    weight = np.ones(len(samples))
    weight[y_samples - x_samples <= 10] = 0.

    prior_mean = op.calculate_mean(weight)
    prior_cov = op.calculate_cov(weight)
    #print('After applying prior that y-x > 10:')
    #print('Mean values of (x,y,z) = ', prior_mean)
    #print('Uncertainties of (x,y,z) = ', np.sqrt(prior_cov.diagonal()))
    #print('Full covariance matrix = \n', prior_cov)

    # Note that the means shifted.  y is pushed up and x is pushed down.
    # Also, x and y are now correlated.
    # And all uncertainties moved down, even z (slightly)

    assert prior_mean[0] < x0
    assert prior_mean[1] > y0
    assert prior_cov[0,1] > 0
    assert all(np.sqrt(prior_cov.diagonal()[:2]) < sigma)
